use std::{fmt, marker::PhantomData};

use rowan::{GreenNode, GreenNodeData, GreenTokenData, NodeOrToken};

use crate::{SyntaxElement, SyntaxError};

pub struct ParserOutput<SyntaxKind: SyntaxElement, SyntaxErr: SyntaxError> {
    pub green_node: GreenNode,
    pub errors: Vec<SyntaxErr>,
    pub(super) _syntax_kind: PhantomData<SyntaxKind>,
}

impl<SyntaxKind: SyntaxElement, SyntaxErr: SyntaxError> std::fmt::Debug
    for ParserOutput<SyntaxKind, SyntaxErr>
{
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let mut errs: Vec<&SyntaxErr> = self.errors.iter().collect();
        errs.reverse();
        debug_print_output::<SyntaxKind, SyntaxErr>(
            NodeOrToken::Node(&self.green_node),
            f,
            0,
            &mut errs,
        )
    }
}

fn debug_print_output<SyntaxKind: SyntaxElement, SyntaxErr: SyntaxError>(
    node: NodeOrToken<&GreenNodeData, &GreenTokenData>,
    f: &mut std::fmt::Formatter<'_>,
    lvl: i32,
    errs: &mut Vec<&SyntaxErr>,
) -> std::fmt::Result {
    for _ in 0..lvl {
        f.write_str("    ")?;
    }

    match node {
        NodeOrToken::Node(n) => {
            let kind: SyntaxKind = node.kind().into();
            if kind != SyntaxKind::SYNTAX_ERROR {
                writeln!(f, "{:?} {{", kind)?;
            } else {
                let err = errs
                    .pop()
                    .expect("all error syntax nodes should correspond to an error");

                writeln!(f, "{:?}: {err:?} {{", kind)?;
            }
            for c in n.children() {
                debug_print_output::<SyntaxKind, SyntaxErr>(c, f, lvl + 1, errs)?;
            }
            for _ in 0..lvl {
                f.write_str("    ")?;
            }
            f.write_str("}\n")
        }
        NodeOrToken::Token(t) => {
            writeln!(
                f,
                "{:?} {:?};",
                Into::<SyntaxKind>::into(t.kind()),
                t.text()
            )
        }
    }
}
